"""Main widget package.

This package holds all files needed for the World Bank Data plugin.
"""

# Category metadata.

# Category icon show in the menu
ICON = "icons/category.svg"

# Background color for category background in menu
# and widget icon background in workflow.
BACKGROUND = "light-blue"

# Location of widget help files.
WIDGET_HELP_PATH = (
    # Used for development.
    # You still need to build help pages using
    # make htmlhelp
    # inside doc folder
    ("{DEVELOP_ROOT}/doc/build/htmlhelp/index.html", None),

    ("http://example.com/my-widget/docs/", "")
)



"""Widget for displaying all indicators."""

import textwrap
import logging
from functools import partial
from functools import lru_cache

import simple_wbd
from PyQt4.QtCore import Qt, QThread, QCoreApplication
from PyQt4 import QtGui
from PyQt4 import QtCore
from Orange.widgets import gui
from Orange.widgets.utils import concurrent

TEXTFILTERROLE = next(gui.OrangeUserRole)
logger = logging.getLogger(__name__)


class MySortFilterProxyModel(QtGui.QSortFilterProxyModel):
    """Filter proxy model.

    This class is used for improving filtering of indicators table.
    """
    # pylint: disable=invalid-name
    # camel case names are false positives because they must be used to
    # override the original QtGui.QSortFilterProxyModel functions

    def __init__(self, parent=None):
        QtGui.QSortFilterProxyModel.__init__(self, parent)
        self._filter_strings = []
        self._cache = {}
        self._cache_fixed = {}
        self._cache_prefix = {}
        self._row_text = {}

        # Create a cached version of _filtered_rows
        self._filtered_rows_cached = lru_cache(100)(self._filtered_rows)

    def setSourceModel(self, model):
        """Set the source model for the filter.
        """
        self._filter_strings = []
        self._cache = {}
        self._cache_fixed = {}
        self._cache_prefix = {}
        self._row_text = {}
        QtGui.QSortFilterProxyModel.setSourceModel(self, model)

    def addFilterFixedString(self, string, invalidate=True):
        """ Add `string` filter to the list of filters. If invalidate is
        True the filter cache will be recomputed.
        """
        self._filter_strings.append(string)
        all_rows = range(self.sourceModel().rowCount())
        row_text = [self.rowFilterText(row) for row in all_rows]
        self._cache[string] = [string in text for text in row_text]
        if invalidate:
            self.updateCached()
            self.invalidateFilter()

    def removeFilterFixedString(self, index=-1, invalidate=True):
        """ Remove the `index`-th filter string. If invalidate is True the
        filter cache will be recomputed.
        """
        string = self._filter_strings.pop(index)
        del self._cache[string]
        if invalidate:
            self.updateCached()
            self.invalidate()

    def setFilterFixedStrings(self, strings):
        """Set a list of string to be the new filters.
        """
        to_remove = set(self._filter_strings) - set(strings)
        to_add = set(strings) - set(self._filter_strings)
        for str_ in to_remove:
            self.removeFilterFixedString(
                self._filter_strings.index(str_),
                invalidate=False)

        for str_ in to_add:
            self.addFilterFixedString(str_, invalidate=False)
        self.updateCached()
        self.invalidate()

    def _filtered_rows(self, filter_strings):
        """Return a dictionary mapping row indexes to True False values.

        .. note:: This helper function is wrapped in the __init__ method.

        """
        all_rows = range(self.sourceModel().rowCount())
        cache = self._cache
        return dict([(row, all([cache[str][row] for str in filter_strings]))
                     for row in all_rows])

    def updateCached(self):
        """Update the combined filter cache.
        """
        self._cache_fixed = self._filtered_rows_cached(
            tuple(sorted(self._filter_strings)))

    def setFilterFixedString(self, string):
        """Should this raise an error? It is not being used.
        """
        QtGui.QSortFilterProxyModel.setFilterFixedString(self, string)

    def rowFilterText(self, row):
        """Return text for `row` to filter on.
        """
        f_role = self.filterRole()
        f_column = self.filterKeyColumn()
        s_model = self.sourceModel()
        data = s_model.data(s_model.index(row, f_column), f_role)
        return str(data)

    def filterAcceptsRow(self, row, _):
        return self._cache_fixed.get(row, True)

    def lessThan(self, left, right):
        """Less comparator for columns."""
        if left.column() == 1 and right.column():
            left_gds = str(left.data(Qt.DisplayRole))
            right_gds = str(right.data(Qt.DisplayRole))
            left_gds = left_gds.lstrip("GDS")
            right_gds = right_gds.lstrip("GDS")
            try:
                return int(left_gds) < int(right_gds)
            except ValueError:
                pass
        return QtGui.QSortFilterProxyModel.lessThan(self, left, right)


class IndicatorsTreeView(QtGui.QTreeView):
    """Tree view widget for displaying all indicators."""

    def __init__(self, parent, main_widget=None):
        super().__init__(parent)
        self._main_widget = main_widget
        self._fetch_task = None
        self._indicator_data = None
        self._api = simple_wbd.IndicatorAPI()
        self.setAlternatingRowColors(True)
        self.setEditTriggers(QtGui.QTreeView.NoEditTriggers)
        self.setRootIsDecorated(False)
        self.setSelectionBehavior(QtGui.QAbstractItemView.SelectRows)
        # if not multi_select:
        # self.setSelectionMode(QtGui.QAbstractItemView.MultiSelection)
        self.setSelectionMode(QtGui.QAbstractItemView.ExtendedSelection)

        linkdelegate = gui.LinkStyledItemDelegate(self)
        self.setItemDelegateForColumn(1, linkdelegate)

        self.setSortingEnabled(True)
        self.setUniformRowHeights(True)
        self.viewport().setMouseTracking(True)

        proxy_model = MySortFilterProxyModel(self)
        self.setModel(proxy_model)
        self.selectionModel().selectionChanged.connect(self._update_selection)
        self.viewport().setMouseTracking(True)

        self._executor = concurrent.ThreadExecutor()
        self.fetch_indicators()

    def fetch_indicators(self):
        """Trigger a background job for fetching a new indicator list."""
        self._main_widget.setBlocking(True)
        self._main_widget.setEnabled(False)
        func = partial(
            self._fetch_indicators,
            concurrent.methodinvoke(
                self._main_widget, "set_progress", (float,))
        )
        self._fetch_task = concurrent.Task(function=func)
        self._fetch_task.finished.connect(self._fetch_indicators_finished)
        self._fetch_task.exceptionReady.connect(self._init_exception)
        self._executor.submit(self._fetch_task)

    def _get_selected_ids(self):
        return [i.data(Qt.DisplayRole) for i in self.selectedIndexes()
                if i.column() == 1]

    def _generate_description(self, indicator_id):
        data = self._indicator_data.get(indicator_id, {})
        return textwrap.dedent(
            """
            <b>ID:</b> {} <br>
            <b>Name</b>: {} <br>
            <b>Source</b>: {} <br>
            <b>Description</b>: {} <br>
            <b>Organization</b>: {} <br>
            <b>Topics</b>:
                <ul>{}</ul>
            <br>
            """.format(
                data.get("id", ""),
                data.get("name", ""),
                data.get("source", {}).get("value", ""),
                data.get("sourceNote", ""),
                data.get("sourceOrganization", ""),
                "\n                 ".join(
                    "<li>{}</li>".format(t["value"])
                    for t in data.get("topics", {})
                    if t.get("value")
                ),
            )
        )

    def _update_selection(self):
        ids = self._get_selected_ids()
        self._main_widget.indicator_selection = ids
        text = "\n\n".join(
            self._generate_description(indicator_id=indicator_id)
            for indicator_id in self._get_selected_ids()
        )
        self._main_widget.indicator_description.clear()
        self._main_widget.indicator_description.setHtml(text)
        self._main_widget.description_box.setTitle(
            "Description ({} {} selected)".format(
                len(ids),
                "indicator" if len(ids) == 1 else "indicators",
            )
        )
        self._main_widget.commit_if()

    @staticmethod
    def _get_link(id_):
        return "http://data.worldbank.org/indicator/{}?view=chart".format(id_)

    def _fetch_indicators(self, progress=lambda val: None):
        """Background task for fetching indicators."""
        progress(0)
        def row_item(display_value, item_values=None):
            """Generate a cell item for a given row."""
            if not item_values:
                item_values = {}
            item = QtGui.QStandardItem()
            item.setData(display_value, Qt.DisplayRole)
            for role, value in item_values.items():
                item.setData(value, role)
            return item

        progress(10)
        filter_ = self._main_widget.basic_indicator_filter()
        data = self._api.get_indicators(filter_=filter_)
        self._indicator_data = {ind["id"]: ind for ind in data}
        progress(70)

        indicators = [[""] + row for row in
                      self._api.get_indicator_list(filter_=filter_)]

        model = QtGui.QStandardItemModel()
        model.setHorizontalHeaderLabels(indicators[0])
        for row in indicators[1:]:
            search_string = " | ".join(row).lower()
            row_data = [row_item(item) for item in row]
            row_data[0].setData(search_string, TEXTFILTERROLE)
            row_data[1].setData(self._get_link(row[1]), gui.LinkRole)
            model.appendRow(row_data)

        progress(100)
        if QThread.currentThread() is not QCoreApplication.instance().thread():
            model.moveToThread(QCoreApplication.instance().thread())
        return model

    def _init_exception(self):
        pass

    def _fetch_indicators_finished(self):
        """Finish handler for fetching indicators.

        This takes the _fetch_indicators result and updates the displayed list
        of indicators.
        """
        assert self.thread() is QtCore.QThread.currentThread()
        if self._fetch_task is None:
            return
        model = self._fetch_task.result()
        model.setParent(self)

        proxy = self.model()
        proxy.setFilterKeyColumn(0)
        proxy.setFilterRole(TEXTFILTERROLE)
        proxy.setFilterCaseSensitivity(False)
        # proxy.setFilterFixedString(self.filterString)

        proxy.setSourceModel(model)
        proxy.sort(1, QtCore.Qt.DescendingOrder)

        # self.progressBarFinished()

        for i in range(3):
            self.resizeColumnToContents(i)

        self.setColumnWidth(
            1, min(self.columnWidth(1), 300))
        self.setColumnWidth(
            2, min(self.columnWidth(2), 400))

        self._main_widget.info_data["Indicators"] = model.rowCount()
        self._main_widget.print_info()

        self._main_widget.setBlocking(False)
        self._main_widget.setEnabled(True)



"""Wrapper for simple_wbd api.

This wrapper adds extra functionality to the IndicatorDataset class, for easier
data manipulation and generating Orange data tables.
"""

import datetime
import time
import logging

import numpy

import Orange
import simple_wbd
from orangecontrib.wbd import countries

logger = logging.getLogger(__name__)


class IndicatorDataset(simple_wbd.IndicatorDataset):
    """Extended indicator dataset.

    This class extends the original indicator dataset by adding as_np_array and
    as_orange_table functions.
    """


    @staticmethod
    def _get_iso_date(date_str):
        """Convert wbd date string into iso format date string.

        Convert date strings such as "2005", "2002Q3" and "1999M7" into iso
        formatted date strings "yyyy-mm-dd"
        """
        try:
            if "Q" in date_str:
                year, quarter = date_str.split("Q")
                return datetime.date(int(year), (int(quarter) * 3) - 2, 1)
            elif "M" in date_str:
                year, month = date_str.split("M")
                return datetime.date(int(year), int(month), 1)
            else:
                return datetime.date(int(date_str), 1, 1).isoformat()
        except ValueError:
            # some dates contain invalid date strings such as
            # "Last Known Value" or "1988-2000" and possible some more. See:
            # http://api.worldbank.org/countries/PRY/indicators/
            #   per_lm_ac.avt_q4_urb?date=1960%3A2016&format=json
            #   &per_page=10000
            # http://api.worldbank.org/countries/all/indicators/
            #   DB_mw_19apprentice?format=json&mrv=10&gapfill=y
            return datetime.date.today().isoformat()

    def _time_series_table(self):
        data = self.as_np_array(time_series=True)

        if not data.any():
            return None

        meta_columns = [[time.mktime(date_.timetuple()) if date_ else None]
                        for date_ in data[1:, 0]]
        data_columns = data[1:, 1:]

        meta_domains = [Orange.data.TimeVariable("Date")]

        colum_domains = [Orange.data.ContinuousVariable(column_name)
                         for column_name in data[0, 1:]]

        logger.debug("Generated Orange table of size: %s", data.shape)

        domain = Orange.data.Domain(colum_domains, metas=meta_domains)
        return Orange.data.Table(domain, data_columns, metas=meta_columns)

    def _country_table(self):
        data = self.as_np_array(add_metadata=True)

        if not data.any():
            return None

        meta_columns = data[1:, :7]
        data_columns = data[1:, 7:]

        meta_domains = [Orange.data.StringVariable(name)
                        for name in data[0, :7]]
        colum_domains = [Orange.data.ContinuousVariable(column_name)
                         for column_name in data[0, 7:]]

        logger.debug("Generated Orange table of size: %s", data.shape)

        domain = Orange.data.Domain(colum_domains, metas=meta_domains)
        return Orange.data.Table(domain, data_columns, metas=meta_columns)

    def as_np_array(self, time_series=False, add_metadata=False, **kwargs):
        """Get a 2D numpy array data representation.

        Args:
            time_series: boolean indicating if the first column should contain
                countries or dates. Different indicators are always in columns
                so this is not a transposition of the matrix.
            add_metadata: This can only be used if time_series is false. If
                set to True, it will append country metadata after the first
                column. Columns added: Region, Admin Region, Income Level,
                latitude, longitude, and Lending type.

        Returns:
            2D numpy array with all indicator data.
        """
        data = numpy.array(self.as_list(time_series=time_series,
                                        add_metadata=add_metadata,
                                        **kwargs))

        # list of column indexes that have at least one non zero value
        filter_ = [ind for ind, col in enumerate(data[1:, :].T) if any(col)]
        return data[:, filter_]

    def as_orange_table(self, time_series=False):
        if time_series:
            return self._time_series_table()
        else:
            return self._country_table()


class IndicatorAPI(simple_wbd.IndicatorAPI):
    """Wrapper for Indicator API to use the extended data set."""

    def __init__(self):
        super().__init__(IndicatorDataset)


class ClimateDataset(simple_wbd.ClimateDataset):
    """Extended climate dataset.

    This class extends the original climate dataset by adding as_np_array and
    as_orange_table functions.
    """


    def as_np_array(self, **kwargs):
        """Get a 2D numpy array data representation.

        Returns:
            2D numpy array with all indicator data.
        """
        data = numpy.array(self.as_list(**kwargs))

        # list of column indexes that have at least one non zero value
        filter_ = [ind for ind, col in enumerate(data[1:, :].T) if any(col)]
        return data[:, filter_]

    def _country_table(self, **kwargs):
        data = self.as_np_array(**kwargs)
        if data.shape[0] < 2:
            return None
        alpha3_map = countries.get_alpha3_map()
        meta_columns = data[1:, :1]
        data_columns = data[1:, 1:]
        for row in meta_columns:
            row[0] = alpha3_map.get(row[0], row[0])

        meta_domains = [Orange.data.StringVariable(name)
                        for name in data[0, :1]]
        colum_domains = [Orange.data.ContinuousVariable(column_name)
                         for column_name in data[0, 1:]]

        logger.debug("Generated Orange table of size: %s", data.shape)

        domain = Orange.data.Domain(colum_domains, metas=meta_domains)
        return Orange.data.Table(domain, data_columns, metas=meta_columns)

    def _time_series_table(self, **kwargs):
        data = self.as_np_array(columns=["country", "type"], use_dates=True,
                                **kwargs)
        if data.shape[0] < 2:
            return None
        alpha3_map = countries.get_alpha3_map()
        meta_columns = [[time.mktime(date_.timetuple()) if date_ else None]
                        for date_ in data[1:, 0]]
        data_columns = data[1:, 1:]
        for row in meta_columns:
            row[0] = alpha3_map.get(row[0], row[0])

        meta_domains = [Orange.data.TimeVariable(name)
                        for name in data[0, :1]]
        colum_domains = [Orange.data.ContinuousVariable(column_name)
                         for column_name in data[0, 1:]]

        logger.debug("Generated Orange table of size: %s", data.shape)

        domain = Orange.data.Domain(colum_domains, metas=meta_domains)
        return Orange.data.Table(domain, data_columns, metas=meta_columns)

    def as_orange_table(self, time_series=False, **kwargs):
        if time_series:
            return self._time_series_table(**kwargs)
        else:
            return self._country_table(**kwargs)


class ClimateAPI(simple_wbd.ClimateAPI):
    """Wrapper for Climate API to use the extended data set."""
    # pylint: disable=too-few-public-methods
    # This is just an extension of the returned dataset. We do not need to add
    # any additional functions here.

    def __init__(self):
        super().__init__(ClimateDataset)



"""Module for generating countries and region lists.

This module has all the helper functions for generating proper structures for
CountryTreeWidget.

The three main exposed functions are:
    get_countries_dict - Used in climate widget.
    get_countries_regions_dict - Used in indicator widget.
    get_alpha3_map - used for changing alpha3 codes to country names.
"""
import copy
import logging
from collections import defaultdict
from collections import OrderedDict

import pycountry
import simple_wbd

logger = logging.getLogger(__name__)

MAPPINGS = {
    # Africa
    "Democratic Republic of the Congo": ("Congo, The Democratic Republic of "
                                         "the"),
    "Ivory Coast": "Côte d'Ivoire",
    "Tanzania": "Tanzania, United Republic of",
    "Iran": "Iran, Islamic Republic of",

    # Asia
    "Macau": "Macao",
    "Northern Cyprus": "Cyprus",
    "South Korea": "Korea, Republic of",
    "North Korea": "Korea, Democratic People's Republic of",
    "Vietnam": "Viet Nam",
    "Palestine": "Palestine, State of",
    "Taiwan": "Taiwan, Province of China",
    "Syria": "Syrian Arab Republic",

    # Europe
    "Macedonia": "Macedonia, Republic of",
    "Moldova": "Moldova, Republic of",
    "Russia": "Russian Federation",
    "Vatican City/Holy See": "Holy See (Vatican City State)",

    # North America
    "Bonaire": "Bonaire, Sint Eustatius and Saba",
    "British Virgin Islands": "Virgin Islands, British",
    "Saint Martin": "Saint Martin (French part)",
    "Sint Maarten": "Sint Maarten (Dutch part)",
    "United States of America": "United States",
    "United States Virgin Islands": "Virgin Islands, U.S.",

    # South America
    "Bolivia": "Bolivia, Plurinational State of",
    "Falkland Islands": "Falkland Islands (Malvinas)",
    "Venezuela": "Venezuela, Bolivarian Republic of",

    # Oceania
    "Pitcairn Islands": "Pitcairn",
    "Micronesia": "Micronesia, Federated States of",

    # Antartica
    "French Southern and Antarctic Lands": "French Southern Territories",
}

COUNTRIES = {
    "Africa": [
        "Algeria",
        "Angola",
        "Benin",
        "Botswana",
        "Burkina Faso",
        "Burundi",
        "Cameroon",
        "Cape Verde",
        "Central African Republic",
        "Chad",
        "Comoros",
        "Congo",
        "Democratic Republic of the Congo",
        "Djibouti",
        "Egypt",
        "Equatorial Guinea",
        "Eritrea",
        "Ethiopia",
        "Gabon",
        "Gambia",
        "Ghana",
        "Guinea",
        "Guinea-Bissau",
        "Ivory Coast",
        "Kenya",
        "Lesotho",
        "Liberia",
        "Libya",
        "Madagascar",
        "Malawi",
        "Mali",
        "Mauritania",
        "Mauritius",
        "Mayotte",
        "Morocco",
        "Mozambique",
        "Namibia",
        "Niger",
        "Nigeria",
        "Réunion",
        "Rwanda",
        # "Sahrawi Arab Democratic Republic",
        "Saint Helena, Ascension and Tristan da Cunha",
        # "São Tomé and Príncipe",
        "Senegal",
        "Seychelles",
        "Sierra Leone",
        "Somalia",
        # "Somaliland",
        "South Africa",
        "South Sudan",
        "Sudan",
        "Swaziland",
        "Tanzania",
        "Togo",
        "Tunisia",
        "Uganda",
        "Zambia",
        "Zimbabwe",
    ],
    "Asia": [
        "China",
        "Christmas Island",
        "Hong Kong",
        "India",
        "Indonesia",
        "Iran",
        "Iraq",
        "Israel",
        "Japan",
        "Jordan",
        "Kazakhstan",
        "Kuwait",
        "Kyrgyzstan",
        # "Laos",
        "Lebanon",
        "Macau",
        "Malaysia",
        "Maldives",
        "Mongolia",
        "Myanmar",
        # "Nagorno-Karabakh",
        "Nepal",
        "Northern Cyprus",
        "North Korea",
        "Oman",
        "Pakistan",
        "Palestine",
        "Philippines",
        "Qatar",
        "Saudi Arabia",
        "Singapore",
        "South Korea",
        # "South Ossetia",
        "Sri Lanka",
        "Syria",
        "Taiwan",
        "Tajikistan",
        "Thailand",
        "Turkey",
        "Turkmenistan",
        "United Arab Emirates",
        "Uzbekistan",
        "Vietnam",
        "Yemen",
    ],
    "Europe": [
        "Åland Islands",
        "Albania",
        "Andorra",
        "Austria",
        "Belarus",
        "Belgium",
        "Bosnia and Herzegovina",
        "Bulgaria",
        "Croatia",
        "Czech Republic",
        "Denmark",
        "Estonia",
        "Faroe Islands",
        "Finland",
        "France",
        "Germany",
        "Gibraltar",
        "Greece",
        "Guernsey",
        "Hungary",
        "Iceland",
        "Ireland",
        "Isle of Man",
        "Italy",
        # "Jan Mayen",  # see Svalbard
        "Jersey",
        # "Kosovo",
        "Latvia",
        "Liechtenstein",
        "Lithuania",
        "Luxembourg",
        "Macedonia",
        "Malta",
        "Moldova",
        "Monaco",
        "Montenegro",
        "Netherlands",
        "Norway",
        "Poland",
        "Portugal",
        "Romania",
        "Russia",
        "San Marino",
        "Serbia",
        "Slovakia",
        "Slovenia",
        "Spain",
        # "Svalbard",
        "Svalbard and Jan Mayen",
        "Sweden",
        "Switzerland",
        # "Transnistria",
        "Ukraine",
        "United Kingdom",
        "Vatican City/Holy See",
    ],
    "NorthAmerica": [
        "Anguilla",
        "Antigua and Barbuda",
        "Aruba",
        "Bahamas",
        "Barbados",
        "Belize",
        "Bermuda",
        "Bonaire",
        "British Virgin Islands",
        "Canada",
        "Cayman Islands",
        # "Clipperton Island",
        "Costa Rica",
        "Cuba",
        "Curaçao",
        "Dominica",
        "Dominican Republic",
        "El Salvador",
        "Greenland",
        "Grenada",
        "Guadeloupe",
        "Guatemala",
        "Haiti",
        "Honduras",
        "Jamaica",
        "Martinique",
        "Mexico",
        "Montserrat",
        # "Navassa Island",
        "Nicaragua",
        "Panama",
        "Puerto Rico",
        # "Saba",
        "Saint Barthélemy",
        "Saint Kitts and Nevis",
        "Saint Lucia",
        "Saint Martin",
        "Saint Pierre and Miquelon",
        "Saint Vincent and the Grenadines",
        # "Sint Eustatius",
        "Sint Maarten",
        "Trinidad and Tobago",
        "Turks and Caicos Islands",
        "United States of America",
        "United States Virgin Islands",
    ],
    "SouthAmerica": [
        "Argentina",
        "Bolivia",
        "Brazil",
        "Chile",
        "Colombia",
        "Ecuador",
        "Falkland Islands",
        "French Guiana",
        "Guyana",
        "Paraguay",
        "Peru",
        "South Georgia and the South Sandwich Islands",
        "Suriname",
        "Uruguay",
        "Venezuela",
    ],
    "Oceania": [
        "American Samoa",
        # "Ashmore and Cartier Islands",
        "Australia",
        # "Baker Island",
        "Cook Islands",
        # "Coral Sea Islands",
        "Fiji",
        "French Polynesia",
        "Guam",
        # "Howland Island",
        # "Jarvis Island",
        # "Johnston Atoll",
        # "Kingman Reef",
        "Kiribati",
        "Marshall Islands",
        "Micronesia",
        # "Midway Atoll",
        "Nauru",
        "New Caledonia",
        "New Zealand",
        "Niue",
        "Norfolk Island",
        "Northern Mariana Islands",
        "Palau",
        # "Palmyra Atoll",
        "Papua New Guinea",
        "Pitcairn Islands",
        "Samoa",
        "Solomon Islands",
        "Tokelau",
        "Tonga",
        "Tuvalu",
        "Vanuatu",
        # "Wake Island",
        "Wallis and Futuna",
    ],
    "Antartica": [
        "Bouvet Island",
        "French Southern and Antarctic Lands",
        "Heard Island and McDonald Islands",
        "South Georgia and the South Sandwich Islands",
    ],
}

DATA_STRUCTURE = [
    ("Aggregates", [
        ("Income Levels", [
            "LIC",  # "Low income"
            "MIC",  # "Middle income",
            "LMC",  # "Lower middle income",
            "UMC",  # "Upper middle income",
            "HIC",  # "High income",
        ]),
        ("Regions", [
            "EAS",  # "East Asia & Pacific (all income levels)",
            "ECS",  # "Europe & Central Asia (all income levels)",
            "LCN",  # "Latin America & Caribbean (all income levels)",
            "MEA",  # "Middle East & North Africa (all income levels)",
            "NAC",  # "North America",
            "SAS",  # "South Asia",
            "SSF",  # "Sub-Saharan Africa (all income levels)",
        ]),
        ("Other", [
            "WLD",  # "World",
            "AFR",  # "Africa",
            "ARB",  # "Arab World",
            ("Low & middle income", [
                "LMY",  # "All low and middle income regions",
                "EAP",  # "East Asia & Pacific (developing only)",
                "ECA",  # "Europe & Central Asia (developing only)",
                "LAC",  # "Latin America & Caribbean (developing only)",
                "MNA",  # "Middle East & North Africa (developing only)",
                "SSA",  # "Sub-Saharan Africa (developing only)",
            ]),
            "EMU",  # other high income areas don't exist anymore
            # ("High income", [
            #     "EMU",  # "Euro area",
            #     "OEC",  # "High income: OECD",
            #     "NOC",  # "High income: nonOECD",
            # ]),
            "CEB",  # "Central Europe and the Baltics",
            "EUU",  # "European Union",
            "FCS",  # "Fragile and conflict affected situations",
            "HPC",  # "Heavily indebted poor countries (HIPC)",
            "IBD",  # "IBRD only",
            "IBT",  # "IDA & IBRD total",
            "IDB",  # "IDA blend",
            "IDX",  # "IDA only",
            "IDA",  # "IDA total",
            "LDC",  # "Least developed countries: UN classification",
            "OED",  # "OECD members",
            ("Small states", [
                "SST",  # "All small states",
                "CSS",  # "Caribbean small states",
                "PSS",  # "Pacific island small states",
                "OSS",  # "Other small states",
            ]),
        ]),
    ]),
    ("Countries", []),
]

RENAME_MAP = {
    "SST": "All small states",
    "EMU": "High Income Euro area",
}


def _order_countries_dict(countries):
    return OrderedDict(sorted(countries.items(), key=lambda x: x[1]["name"]))


def get_alpha3_map():
    """Get mappings from alpha3 codes to country names."""
    name_map = {v: k for k, v in MAPPINGS.items()}
    return {c.alpha3: name_map.get(c.name, c.name)
            for c in pycountry.countries}


def get_countries_dict():
    """Get a dict of all all country codes and names grouped by continent."""
    result = defaultdict(dict)
    alpha_map = {c.name: c.alpha3 for c in pycountry.countries}
    for continent, countries in COUNTRIES.items():
        for country in countries:
            result[continent][alpha_map[MAPPINGS.get(country, country)]] = {
                "name": country
            }
    result = {k: _order_countries_dict(v) for k, v in result.items()}
    ordered_result = OrderedDict(sorted(result.items(), key=lambda t: t[0]))
    return ordered_result


def _gather_used_ids(items):
    ids = set()
    for item in items:
        if isinstance(item, str):
            ids.add(item)
        else:
            ids |= _gather_used_ids(item[1])
    return ids


def _add_missing_aggregates(data):
    """Add any missing aggregates to data.

    The data structure might be missing any newly added aggregates from the
    API. This function will add all of those under the other section of
    aggregates list.
    """
    api = simple_wbd.IndicatorAPI()
    countries = api.get_countries()
    aggregate_codes = set(
        [i.get("id") for i in countries if
         i.get("region", {}).get("value") == "Aggregates"]
    )
    used_codes = _gather_used_ids(data)
    missing_aggregates = aggregate_codes.difference(used_codes)
    for code in sorted(missing_aggregates):
        # 0,1 = Aggregates, 2,1 = Other
        data[0][1][2][1].append(code)
    return data


def _add_missing_countries(data):
    """Add all countries from the API to country data."""
    api = simple_wbd.IndicatorAPI()
    countries = api.get_countries()
    country_codes = set(
        [i.get("id") for i in countries if
         i.get("region", {}).get("value") != "Aggregates"]
    )
    for code in sorted(country_codes):
        # 1,1 = Countries
        data[1][1].append(code)
    return data


def _generate_country_map():
    api = simple_wbd.IndicatorAPI()
    countries = api.get_countries()
    return {c["id"]: c for c in countries}


def _generate_country_dict(data):
    """Turn the country and region structure into an ordered dict."""
    country_map = _generate_country_map()
    country_dict = OrderedDict()

    for item in data:
        if isinstance(item, tuple):
            country_dict[item[0]] = _generate_country_dict(item[1])
        elif item in country_map:
            country_dict[item] = country_map[item]
        else:
            logger.info("Missing country item: %s", item)
    return country_dict


def get_countries_regions_dict():
    """Get country and region data for indicators widget."""
    data = copy.deepcopy(DATA_STRUCTURE)
    data = _add_missing_aggregates(data)
    data = _add_missing_countries(data)
    return _generate_country_dict(data)



"""Model with the main Orange Widget.

This module contains the world bank data widget, used for fetching data from
world bank data API.
"""

import logging
import collections
from functools import partial

import requests
from PyQt4 import QtCore
from Orange.widgets import widget
from Orange.widgets.utils import concurrent


logger = logging.getLogger(__name__)


class OWWidgetBase(widget.OWWidget):
    """Base class for Indicators and Climate widgets."""

    # pylint: disable=invalid-name
    # Some names have to be invalid to override parent fields.
    # pylint: disable=too-many-ancestors
    # False positive from fetching all ancestors from QWWidget.
    # pylint: disable=too-many-instance-attributes
    # False positive from fetching all attributes from QWWidget.

    category = "Data Sets"
    country_selection = None

    def __init__(self):
        super().__init__()
        logger.debug("Initializing %s", self.__class__.__name__)
        self._fetch_task = None
        self._info_label = None
        self._selection_changed = False
        self._set_progress_flag = False
        self._executor = concurrent.ThreadExecutor()
        self.info_data = collections.OrderedDict([
            ("Server status", None),
            ("Indicators", None),
            ("Selected countries", None),
            ("Rows", None),
            ("Columns", None),
            ("Warning", None),
        ])

    def _check_server_status(self):
        try:
            requests.get('http://api.worldbank.org', timeout=1)
            self.info_data["Server status"] = "Up"
        except requests.exceptions.Timeout:
            self.info_data["Server status"] = "Down"
        self.print_info()

    def print_info(self):
        """Refresh info in the info label."""
        if not self._info_label:
            return
        lines = ["{}: {}".format(k, v) for k, v in self.info_data.items()
                 if v is not None]
        self._info_label.setText("\n".join(lines))

    def commit_if(self):
        """Auto commit handler.

        This function must be called on every action that should trigger an
        auto commit.
        """
        logger.debug("Commit If - auto_commit: %s", self.auto_commit)
        if self.auto_commit:
            self.commit()
        else:
            self._selection_changed = True

    def commit(self):
        """Fetch the climate data and send a new orange table."""
        logger.debug("commit data")
        self.setEnabled(False)
        self._set_progress_flag = True

        func = partial(
            self._fetch_dataset,
            concurrent.methodinvoke(self, "set_progress", (float,))
        )
        self._fetch_task = concurrent.Task(function=func)
        self._fetch_task.finished.connect(self._fetch_dataset_finished)
        self._fetch_task.exceptionReady.connect(self._fetch_dataset_exception)
        self._executor.submit(self._fetch_task)

    def _fetch_dataset(self, set_progress=None):
        raise NotImplementedError(
            "Missing implementation for _fetch_dataset.")

    def _dataset_to_table(self, dataset):
        raise NotImplementedError(
            "Missing implementation for _dataset_to_table.")

    def _fetch_dataset_finished(self):
        """Send data signal on finished dataset fetch."""
        assert self.thread() is QtCore.QThread.currentThread()
        self.setEnabled(True)
        self._set_progress_flag = False
        self.set_progress(100)

        if self._fetch_task is None:
            return

        dataset = self._fetch_task.result()
        data_table = self._dataset_to_table(dataset)

        self.print_info()
        self.send("Data", data_table)

    @staticmethod
    def _fetch_dataset_exception(exception):
        logger.exception(exception)

    @QtCore.pyqtSlot(float)
    def set_progress(self, value):
        """set widgets progress indicator.

        Args:
            value: integer indicating number of percent finished.
        """
        # pylint: disable=invalid-name
        # the progressBarValue is defined in a super class and can not be
        # changed here.
        logger.debug("Set progress: %s", value)
        self.progressBarValue = value
        if value == 100:
            self.progressBarFinished()

    def get_country_codes(self):
        """Get a list of alpha3 codes for selected countries or regions."""
        if self.country_selection:
            # pylint: disable=no-member
            # Settings instance can have items member if it is defined as dict.
            return [k for k, v in self.country_selection.items()
                    if v == 2 and len(str(k)) == 3]
        return []

    def _dataset_progress(self, set_progress=None):
        pass

    def _start_progerss_task(self):
        func = partial(
            self._dataset_progress,
            concurrent.methodinvoke(self, "set_progress", (float,))
        )
        progress_task = concurrent.Task(function=func)
        progress_task.exceptionReady.connect(self._dataset_progress_exception)
        self._executor.submit(progress_task)



"""Model with the main Orange Widget.

This module contains the world bank data widget, used for fetching data from
world bank data API.
"""

import sys
import time
import math
import signal
import logging

from PyQt4 import QtGui
from Orange.data import table
from Orange.widgets import widget
from Orange.widgets import gui
from Orange.widgets.settings import Setting

from orangecontrib.wbd.countries_and_regions import CountryTreeWidget
from orangecontrib.wbd import api_wrapper
from orangecontrib.wbd import countries
from orangecontrib.wbd import owwidget_base

logger = logging.getLogger(__name__)


class OWWorldBankClimate(owwidget_base.OWWidgetBase):
    """World bank data widget for Orange."""
    # pylint: disable=invalid-name
    # Some names have to be invalid to override parent fields.
    # pylint: disable=too-many-ancestors
    # False positive from fetching all ancestors from QWWidget.
    # pylint: disable=too-many-instance-attributes
    # False positive from fetching all attributes from QWWidget.

    # Widget needs a name, or it is considered an abstract widget
    # and not shown in the menu.
    name = "WB Climate"
    icon = "icons/climate.png"
    outputs = [widget.OutputSignal(
        "Data",
        table.Table,
        doc="Climate data from World bank Climate API"
    )]

    settingsList = [
        "auto_commit",
        "country_selection",
        "mergeSpots",
        "output_type"
        "splitterSettings",
    ]

    country_selection = Setting({})
    output_type = Setting(True)
    mergeSpots = Setting(True)
    auto_commit = Setting(False)
    use_country_names = Setting(False)

    include_intervals = Setting([])
    include_data_types = Setting([])

    def _data_type_setter(self, name, value):
        intervals = set(self.include_data_types) | {name}
        if not value:
            intervals.remove(name)
        self.include_data_types = list(intervals)
        logger.debug("New intervals: %s", self.include_data_types)

    def _interval_setter(self, name, value):
        intervals = set(self.include_intervals) | {name}
        if not value:
            intervals.remove(name)
        self.include_intervals = list(intervals)
        logger.debug("New intervals: %s", self.include_intervals)

    @property
    def include_month(self):
        return "month" in self.include_intervals

    @include_month.setter
    def include_month(self, value):
        self._interval_setter("month", value)

    @property
    def include_year(self):
        return "year" in self.include_intervals

    @include_year.setter
    def include_year(self, value):
        self._interval_setter("year", value)

    @property
    def include_decade(self):
        return "decade" in self.include_intervals

    @include_decade.setter
    def include_decade(self, value):
        self._interval_setter("decade", value)

    @property
    def include_temperature(self):
        return "tas" in self.include_data_types

    @include_temperature.setter
    def include_temperature(self, value):
        self._data_type_setter("tas", value)

    @property
    def include_precipitation(self):
        return "pr" in self.include_data_types

    @include_precipitation.setter
    def include_precipitation(self, value):
        self._data_type_setter("pr", value)

    splitterSettings = Setting((
        b'\x00\x00\x00\xff\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00\x01\xea\x00'
        b'\x00\x00\xd7\x01\x00\x00\x00\x07\x01\x00\x00\x00\x02',
        b'\x00\x00\x00\xff\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00\x01\xb5\x00'
        b'\x00\x02\x10\x01\x00\x00\x00\x07\x01\x00\x00\x00\x01'
    ))

    def __init__(self):
        super().__init__()
        self._api = api_wrapper.ClimateAPI()
        self._init_layout()
        self.print_selection_count()
        self._check_server_status()

    def print_selection_count(self):
        """Update info widget with new selection count."""
        country_codes = self.get_country_codes()
        self.info_data["Selected countries"] = len(country_codes)
        self.print_info()

    def _init_layout(self):
        """Initialize widget layout."""

        # Control area
        info_box = gui.widgetBox(self.controlArea, "Info", addSpace=True)
        self._info_label = gui.widgetLabel(info_box, "Initializing\n\n")

        box = gui.vBox(self.controlArea, "Average intervals:")
        self.ch_month = gui.checkBox(box, self, "include_month", "Month",
                                     callback=self.commit_if)
        self.ch_year = gui.checkBox(box, self, "include_year", 'Year',
                                    callback=self.commit_if)
        self.ch_decade = gui.checkBox(box, self, "include_decade", 'Decade',
                                      callback=self.commit_if)

        box = gui.vBox(self.controlArea, "Data Types")
        gui.checkBox(box, self, "include_temperature", "Temperature",
                     callback=self.commit_if)
        gui.checkBox(box, self, "include_precipitation", 'Precipitation',
                     callback=self.commit_if)

        output_box = gui.widgetBox(self.controlArea, "Output", addSpace=True)
        gui.radioButtonsInBox(output_box, self, "output_type",
                              ["Countries", "Time Series"], "Rows",
                              callback=self.output_type_selected)

        gui.checkBox(output_box, self, "use_country_names",
                     "Use Country names", callback=self.commit_if)
        self.output_type = 0

        # pylint: disable=duplicate-code
        gui.separator(output_box)
        gui.auto_commit(self.controlArea, self, "auto_commit", "Commit",
                        box="Commit")
        gui.rubber(self.controlArea)

        # Main area
        box = gui.widgetBox(self.mainArea, "Countries")
        self.country_tree = CountryTreeWidget(
            self.mainArea,
            self.country_selection,
            commit_callback=self.commit_if,
            default_colapse=True,
            default_select=False,
        )
        self.country_tree.set_data(countries.get_countries_dict())
        box.layout().addWidget(self.country_tree)
        self.resize(500, 400)  # why does this not work

    def output_type_selected(self):
        """Output type handle."""
        logger.debug("output type set to: %s", self.output_type)
        if self.output_type == 1:  # Time series
            self.ch_decade.setEnabled(False)
            self.ch_month.setEnabled(False)
            self.ch_year.setEnabled(False)
            self.include_year = True
            self.include_month = False
            self.include_decade = False
        else:
            self.ch_decade.setEnabled(True)
            self.ch_month.setEnabled(True)
            self.ch_year.setEnabled(True)
        self.commit_if()

    def _splitter_moved(self, *_):
        self.splitterSettings = [bytes(sp.saveState())
                                 for sp in self.splitters]

    def _check_big_selection(self):
        types = len(self.include_data_types) if self.include_data_types else 2
        intervals = len(
            self.include_intervals) if self.include_intervals else 2
        country_codes = self.get_country_codes()
        selected_countries = len(country_codes)
        if types * intervals * selected_countries > 100:
            self.info_data[
                "Warning"] = "Fetching data\nmight take a few minutes."
        else:
            self.info_data["Warning"] = None
        self.print_info()

    def commit_if(self):
        """Auto commit handler.

        This function must be called on every action that should trigger an
        auto commit.
        """
        self._check_big_selection()
        self.print_selection_count()
        super().commit_if()

    def _fetch_dataset(self, set_progress=None):
        """Fetch climate dataset."""

        set_progress(0)
        self._start_progerss_task()

        country_codes = self.get_country_codes()

        logger.debug("Fetch: selected country codes: %s", country_codes)
        climate_dataset = self._api.get_instrumental(
            country_codes,
            data_types=self.include_data_types,
            intervals=self.include_intervals
        )
        self._set_progress_flag = False
        return climate_dataset

    def _dataset_to_table(self, dataset):
        time_series = self.output_type == 1
        return dataset.as_orange_table(
            time_series=time_series,
            use_names=self.use_country_names,
        )

    def _dataset_progress(self, set_progress=None):
        while self._set_progress_flag:
            pages = self._api.progress["pages"]
            current_page = self._api.progress["current_page"]
            logger.debug("api progress: %s", self._api.progress)
            if pages > 0:
                progress = ((100 / pages) * (current_page - 1))
                logger.debug("calculated progress: %s", progress)
                set_progress(math.floor(progress))
            time.sleep(1)

    @staticmethod
    def _dataset_progress_exception(exception):
        logger.exception(exception)


def main():  # pragma: no cover
    """Helper for running the widget without Orange."""
    logging.basicConfig(level=logging.DEBUG)
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    app = QtGui.QApplication(sys.argv)
    climate_widget = OWWorldBankClimate()
    climate_widget.show()
    app.exec_()
    climate_widget.saveSettings()


if __name__ == "__main__":
    main()



"""Module contains countries and regions tree widget."""

import collections
import logging

import simple_wbd
from PyQt4 import QtGui
from PyQt4 import QtCore

from orangecontrib.wbd import countries

logger = logging.getLogger(__name__)


class CountryTreeWidget(QtGui.QTreeWidget):
    """Display a list of countries and their codes.

    The aggregates list was taken from:
        https://datahelpdesk.worldbank.org/knowledgebase/articles/898614-api-aggregates-regions-and-income-levels
    Groups were taken from:
        https://datahelpdesk.worldbank.org/knowledgebase/articles/906519
    """

    def __init__(self, parent, selection_list, commit_callback=None,
                 default_colapse=False, default_select=False):
        super().__init__(parent)
        if default_select:
            self._default_select = QtCore.Qt.Checked
        else:
            self._default_select = QtCore.Qt.Unchecked
        self._default_colapse = default_colapse
        self._commit_callback = commit_callback
        self._selection_list = selection_list
        self._busy = False
        self._api = simple_wbd.IndicatorAPI()
        self._init_view()
        self._init_listeners()

    def _init_listeners(self):
        self.itemChanged.connect(self.selection_changed)
        self.itemCollapsed.connect(self._item_collapsed)
        self.itemExpanded.connect(self._item_expanded)

    def _item_expanded(self, item):
        self._selection_list[("collapsed", item.key)] = False

    def _item_collapsed(self, item):
        self._selection_list[("collapsed", item.key)] = True

    def _init_view(self):
        self.setHeaderLabels(["Regions and countries", "Code"])
        self.setRootIsDecorated(True)

    def selection_changed(self, item, _):
        """Country selection change handler.

        This function updates the settings dict with the new selection and
        triggers commit on change callback.

        Args:
            item: list item that has changed.
        """
        if self._busy:
            return
        state = item.checkState(0)
        self._selection_list[item.key] = state
        logger.debug("Selection for item '%s' set to %s", item.key, state)
        if self._commit_callback:
            self._commit_callback()

    def _fill_values(self, data, parent=None):
        if not parent:
            parent = self

        tristate = QtCore.Qt.ItemIsUserCheckable | QtCore.Qt.ItemIsTristate
        twostate = QtCore.Qt.ItemIsUserCheckable
        defaults = self._selection_list
        for key, value in data.items():
            name = countries.RENAME_MAP.get(key, value.get("name", key))
            display_key = "" if name == key else key

            item = QtGui.QTreeWidgetItem(parent, [name, display_key])
            item.key = value if isinstance(value, str) else key

            defaults[item.key] = defaults.get(item.key, self._default_select)
            item.setCheckState(0, defaults[item.key])
            if isinstance(value, collections.OrderedDict):
                item.setFlags(item.flags() | tristate)
                self._fill_values(value, item)
            else:
                item.setFlags(item.flags() | twostate)

    def _collapse_items(self, root=None):
        """Collapse items that were marked as collapsed in previous sessions.

        Args:
            root: Item whose children we want to check and collapse. If none is
                set, root item is chosen and we iterate through all top level
                items.
        """
        if root is None:
            root = self.invisibleRootItem()
        child_count = root.childCount()
        for i in range(child_count):
            item = root.child(i)
            if not item.childCount():
                continue
            if self._selection_list.get(("collapsed", item.key), self._default_colapse):
                item.setExpanded(False)
            self._collapse_items(item)

    def set_data(self, data):
        """Fill the country and region list with new data.

        Args:
            data: dict of dictionaries. The last dictionary must contain a
                field "name" which is displayed in the list.
        """
        self._busy = True
        self.clear()
        self._fill_values(data)
        self.expandAll()
        self._collapse_items()

        for i in range(self.columnCount()):
            self.resizeColumnToContents(i)
        self._busy = False



#!/usr/bin/env python

from setuptools import setup

ENTRY_POINTS = {
    # Entry point used to specify packages containing tutorials accessible
    # from welcome screen. Tutorials are saved Orange Workflows (.ows files).
    'orange.widgets.tutorials': (
        # Syntax: any_text = path.to.package.containing.tutorials
        'exampletutorials = orangecontrib.wbd.tutorials',
    ),

    # Entry point used to specify packages containing widgets.
    'orange.widgets': (
        # Syntax: category name = path.to.package.containing.widgets
        # Widget category specification can be seen in
        #    orangecontrib/wbd/widgets/__init__.py
        'Data Sets = orangecontrib.wbd.widgets',
    ),

    # Register widget help
    "orange.canvas.help": (
        'html-index = orangecontrib.wbd.widgets:WIDGET_HELP_PATH',)
}

KEYWORDS = [
    # [PyPi](https://pypi.python.org) packages with keyword "orange3 add-on"
    # can be installed using the Orange Add-on Manager
    "orange3 add-on",
    "world bank data",
    "indicator api",
]


def get_description():
    with open("README.rst") as f:
        return f.read()


if __name__ == '__main__':
    setup(
        name="Orange3-Datasets",
        version="0.1.3",
        license="MIT",
        author="Miha Zidar",
        author_email="zidarsk8@gmail.com",
        description=("Orange interface for World Bank Data Indicator and "
                     "Climate APIs"),
        long_description=get_description(),
        url="https://github.com/zidarsk8/orange3-datasets",
        download_url=("https://github.com/zidarsk8/orange3-datasets/tarball/"
                      "0.1.3"),
        packages=[
            'orangecontrib',
            'orangecontrib.wbd',
            'orangecontrib.wbd.tutorials',
            'orangecontrib.wbd.widgets',
        ],
        package_data={
            'orangecontrib.wbd': ['tutorials/*.ows'],
            'orangecontrib.wbd.widgets': ['icons/*'],
        },
        install_requires=[
            'Orange3',
            'numpy',
            'observable',
            'simple_wbd>=0.5.1',
        ],
        extras_require={
            'doc': [
                'sphinx',
                'sphinx_rtd_theme',
            ],
            'test': [
                'nose',
                'nose-cov',
            ]
        },
        entry_points=ENTRY_POINTS,
        keywords=KEYWORDS,
        namespace_packages=['orangecontrib'],
        include_package_data=True,
        zip_safe=False,
        classifiers=[
            'Development Status :: 4 - Beta',
            'Environment :: X11 Applications :: Qt',
            'Environment :: Plugins',
            'Programming Language :: Python',
            'Operating System :: OS Independent',
            'Topic :: Scientific/Engineering :: Artificial Intelligence',
            'Topic :: Scientific/Engineering :: Visualization',
            'Topic :: Software Development :: Libraries :: Python Modules',
            'Intended Audience :: Education',
            'Intended Audience :: Science/Research',
            'Intended Audience :: Developers',
        ],
    )



"""Model with the main Orange Widget.

This module contains the world bank data widget, used for fetching data from
world bank data API.
"""

import sys
import time
import math
import signal
import logging
import collections

from PyQt4 import QtGui
from PyQt4 import QtCore
from Orange.data import table
from Orange.widgets import widget
from Orange.widgets import gui
from Orange.widgets.settings import Setting

from orangecontrib.wbd.countries_and_regions import CountryTreeWidget
from orangecontrib.wbd.indicators_list import IndicatorsTreeView
from orangecontrib.wbd import api_wrapper
from orangecontrib.wbd import countries
from orangecontrib.wbd import owwidget_base

logger = logging.getLogger(__name__)


class OWWorldBankIndicators(owwidget_base.OWWidgetBase):
    """World bank data widget for Orange."""
    # pylint: disable=invalid-name
    # Some names have to be invalid to override parent fields.
    # pylint: disable=too-many-ancestors
    # False positive from fetching all ancestors from QWWidget.
    # pylint: disable=too-many-instance-attributes
    # False positive from fetching all attributes from QWWidget.

    # Widget needs a name, or it is considered an abstract widget
    # and not shown in the menu.
    name = "WB Indicators"
    icon = "icons/wb_icon.png"
    outputs = [widget.OutputSignal(
        "Data",
        table.Table,
        doc="Indicator data from World bank Indicator API"
    )]

    indicator_list_map = collections.OrderedDict([
        (0, "All"),
        (1, "Common"),
        (2, "Featured"),
    ])

    settingsList = [
        "indicator_list_selection",
        "country_selection",
        "indicator_selection",
        "splitterSettings",
        "currentGds",
        "auto_commit",
        "output_type",
    ]

    country_selection = Setting({})
    indicator_selection = Setting([])
    indicator_list_selection = Setting(True)
    output_type = Setting(True)
    auto_commit = Setting(False)

    splitterSettings = Setting((
        b'\x00\x00\x00\xff\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00\x01\xea'
        b'\x00\x00\x00\xd7\x01\x00\x00\x00\x07\x01\x00\x00\x00\x02',
        b'\x00\x00\x00\xff\x00\x00\x00\x00\x00\x00\x00\x02\x00\x00\x01\xb5'
        b'\x00\x00\x02\x10\x01\x00\x00\x00\x07\x01\x00\x00\x00\x01'
    ))

    def __init__(self):
        super().__init__()
        self._api = api_wrapper.IndicatorAPI()
        self._init_layout()
        self._check_server_status()

    def _init_layout(self):
        """Initialize widget layout."""

        # Control area
        info_box = gui.widgetBox(self.controlArea, "Info", addSpace=True)
        self._info_label = gui.widgetLabel(info_box, "Initializing\n\n")

        indicator_filter_box = gui.widgetBox(self.controlArea, "Indicators",
                                             addSpace=True)
        gui.radioButtonsInBox(indicator_filter_box, self,
                              "indicator_list_selection",
                              self.indicator_list_map.values(), "Rows",
                              callback=self.indicator_list_selected)
        self.indicator_list_selection = 2

        gui.separator(indicator_filter_box)

        output_box = gui.widgetBox(self.controlArea, "Output", addSpace=True)
        gui.radioButtonsInBox(output_box, self, "output_type",
                              ["Countries", "Time Series"], "Rows",
                              callback=self.output_type_selected)
        self.output_type = 0

        gui.separator(output_box)
        # pylint: disable=duplicate-code
        gui.auto_commit(self.controlArea, self, "auto_commit", "Commit",
                        box="Commit")
        gui.rubber(self.controlArea)

        # Main area

        gui.widgetLabel(self.mainArea, "Filter")
        self.filter_text = QtGui.QLineEdit(
            textChanged=self.filter_indicator_list)
        self.completer = QtGui.QCompleter(
            caseSensitivity=QtCore.Qt.CaseInsensitive)
        self.completer.setModel(QtGui.QStringListModel(self))
        self.filter_text.setCompleter(self.completer)

        spliter_v = QtGui.QSplitter(QtCore.Qt.Vertical, self.mainArea)

        self.mainArea.layout().addWidget(self.filter_text)
        self.mainArea.layout().addWidget(spliter_v)

        self.indicator_widget = IndicatorsTreeView(spliter_v, main_widget=self)

        splitter_h = QtGui.QSplitter(QtCore.Qt.Horizontal, spliter_v)

        self.description_box = gui.widgetBox(splitter_h, "Description")

        self.indicator_description = QtGui.QTextEdit()
        self.indicator_description.setReadOnly(True)
        self.description_box.layout().addWidget(self.indicator_description)

        box = gui.widgetBox(splitter_h, "Countries and Regions")
        self.country_tree = CountryTreeWidget(
            splitter_h,
            self.country_selection,
            default_select=True,
            default_colapse=True,
        )
        box.layout().addWidget(self.country_tree)
        self.country_tree.set_data(countries.get_countries_regions_dict())

        self.splitters = spliter_v, splitter_h

        for splitter, setting in zip(self.splitters, self.splitterSettings):
            splitter.splitterMoved.connect(self._splitter_moved)
            splitter.restoreState(setting)

        # self.resize(2000, 600)  # why does this not work

        self.progressBarInit()

    def filter_indicator_list(self):
        """Set the proxy model filter and update info box."""
        filter_string = self.filter_text.text()
        proxy_model = self.indicator_widget.model()
        if proxy_model:
            strings = filter_string.lower().strip().split()
            proxy_model.setFilterFixedStrings(strings)
            self.print_info()

    def output_type_selected(self):
        self.commit_if()

    def basic_indicator_filter(self):
        return self.indicator_list_map.get(self.indicator_list_selection)

    def indicator_list_selected(self):
        """Update basic indicator selection.

        Switch indicator list selection between All, Common, and Featured.
        """
        value = self.basic_indicator_filter()
        logger.debug("Indicator list selected: %s", value)
        self.indicator_widget.fetch_indicators()

    def _splitter_moved(self, *_):
        self.splitterSettings = [bytes(sp.saveState())
                                 for sp in self.splitters]

    def _fetch_dataset(self, set_progress=None):
        """Fetch indicator dataset."""
        set_progress(0)
        self._start_progerss_task()

        country_codes = self.get_country_codes()

        if len(country_codes) > 250:
            country_codes = None
        logger.debug("Fetch: selected country codes: %s", country_codes)
        logger.debug("Fetch: selected indicators: %s",
                     self.indicator_selection)
        indicator_dataset = self._api.get_dataset(self.indicator_selection,
                                                  countries=country_codes)
        self._set_progress_flag = False
        return indicator_dataset

    def _dataset_to_table(self, dataset):
        time_series = self.output_type == 1
        return dataset.as_orange_table(time_series=time_series)

    @staticmethod
    def _fetch_dataset_exception(exception):
        logger.exception(exception)

    def _dataset_progress(self, set_progress=None):
        """Update dataset download progress.

        This function reads the progress state from the world bank API and sets
        the current widgets progress to that. All This thread should only read
        data and ask the GUI thread to update the progress for this to be
        thread safe.
        """
        while self._set_progress_flag:
            indicators = self._api.progress["indicators"]
            current_indicator = self._api.progress["current_indicator"]
            indicator_pages = self._api.progress["indicator_pages"]
            current_page = self._api.progress["current_page"]
            logger.debug("api progress: %s", self._api.progress)
            if indicator_pages > 0 and indicators > 0:
                progress = (
                    ((100 / indicators) * (current_indicator - 1)) +
                    (100 / indicators) * (current_page / indicator_pages)
                )
                logger.debug("calculated progress: %s", progress)
                set_progress(math.floor(progress))
            time.sleep(1)

    def _dataset_progress_exception(self, exception):
        logger.exception(exception)
        self.print_info()


def main():  # pragma: no cover
    """Helper for running the widget without Orange."""
    logging.basicConfig(level=logging.DEBUG)
    signal.signal(signal.SIGINT, signal.SIG_DFL)
    app = QtGui.QApplication(sys.argv)
    indicators_widget = OWWorldBankIndicators()
    indicators_widget.show()
    app.exec_()
    indicators_widget.saveSettings()


if __name__ == "__main__":
    main()



"""Default package name for orange plugins.

orangecontrib is a namespace modules shared by multiple Orange add-on so it
needs to declare namespace.
"""

__import__("pkg_resources").declare_namespace(__name__)
